# -*- coding: utf-8 -*-

"""Meepo2 Replicators based on events.
"""

from __future__ import absolute_import

__all__ = ["QueueReplicator", "RqReplicator"]

import logging
import zmq

zmq_ctx = zmq.Context()


class Replicator(object):
    """Replicator base class.
    """
    def __init__(self, listen=None, name="meepo2.replicator.zmq"):
        """
        :param listen: zeromq dsn to connect, can be a list
        """
        # replicator logger naming
        self.name = name
        self.logger = logging.getLogger(name)

        self.listen = listen
        self.socket = zmq_ctx.socket(zmq.SUB)

    def run(self):
        raise NotImplementedError()

    def event(self):
        raise NotImplementedError()


from .queue import QueueReplicator
from .rq import RqReplicator



# -*- coding: utf-8 -*-

from __future__ import absolute_import

from ._compat import str, bytes


def _monkey_patch_hashable_func():
    def hashable_identity(obj):
        if hasattr(obj, '__func__'):
            return (id(obj.__func__), id(obj.__self__))
        elif hasattr(obj, 'im_func'):
            return (id(obj.im_func), id(obj.im_self))
        elif isinstance(obj, (str, bytes)):
            return obj
        # hack for session hash info
        elif hasattr(obj, "info") and obj.info:
            return hash(str(sorted(obj.info.items())))
        else:
            return id(obj)

    import blinker.base
    blinker.base.hashable_identity = hashable_identity


_monkey_patch_hashable_func()


from blinker import Namespace

# The namespace for code signals.  If you are not flask code, do
# not put signals in here.  Create your own namespace instead.
_signals = Namespace()
signal = _signals.signal



# -*- coding: utf-8 -*-

from __future__ import absolute_import

import itertools
import logging

from ..signals import signal


def zmq_sub(bind, tables, forwarder=False, green=False):
    """0mq fanout sub.

    This sub will use zeromq to fanout the events.

    :param bind: the zmq pub socket or zmq device socket.
    :param tables: the events of tables to follow.
    :param forwarder: set to True if zmq pub to a forwarder device.
    :param green: weather to use a greenlet compat zmq
    """
    logger = logging.getLogger("meepo2.sub.zmq_sub")

    if not isinstance(tables, (list, set)):
        raise ValueError("tables should be list or set")

    if not green:
        import zmq
    else:
        import zmq.green as zmq

    ctx = zmq.Context()
    socket = ctx.socket(zmq.PUB)

    if forwarder:
        socket.connect(bind)
    else:
        socket.bind(bind)

    events = ("%s_%s" % (tb, action) for tb, action in
              itertools.product(*[tables, ["write", "update", "delete"]]))
    for event in events:
        def _sub(pk, event=event):
            msg = "%s %s" % (event, pk)
            socket.send_string(msg)
            logger.debug("pub msg: %s" % msg)
        signal(event).connect(_sub, weak=False)

    return socket



# -*- coding: utf-8 -*-

"""
Prepare Commit also known as Two-Phase Commit, for basic concept about it,
refer to http://en.wikipedia.org/wiki/Two-phase_commit_protocol


The two phase commit feature implemented in meepo2 is used to make sure event
100% reliably recorded in eventsourcing, and it's not a strict traditional
two-phase commit.

Only use it if you need a 100% grantee of not losing any events. The feature
should only be used in combination of sqlalchemy_es_pub, which ships with
session prepare-commit signals.
"""

from __future__ import absolute_import

import functools
import logging
import pickle
import time

import redis

from ...utils import d, s


class PrepareCommit(object):
    """Prepare-Commit base class, defines the essential APIs.
    """
    def __init__(self):
        pass

    def prepare(self):
        raise NotImplementedError

    def commit(self):
        raise NotImplementedError

    def rollback(self):
        pass


def _redis_strict_pc(func):
    """Strict deco for RedisPrepareCommit

    The deco will choose whether to silent exception or not based on the
    strict attr in RedisPrepareCommit object.
    """
    phase = "session_%s" % func.__name__

    @functools.wraps(func)
    def wrapper(self, session, *args, **kwargs):
        try:
            func(self, session, *args, **kwargs)
            self.logger.debug("%s -> %s" % (session.meepo2_unique_id, phase))
            return True
        except Exception as e:
            if self.strict:
                raise
            if isinstance(e, redis.ConnectionError):
                self.logger.warn("redis connection error in %s: %s" % (
                    phase, session.meepo2_unique_id))
            else:
                self.logger.exception(e)
            return False
    return wrapper


class RedisPrepareCommit(PrepareCommit):
    """Prepare Commit session based on redis.

    This prepare commit records sqlalchemy session, and should be used with
    :func:`sqlalchemy_es_pub`.

    :param redis_dsn: the redis instance uri
    :param strict: by default the exceptions happened in middle of
     prepare-commit will only be caught and logged as error, but the
     process continue to execute. If strict set to True, the exception
     will be  raised to outside.
    :param namespace: namespace string or namespace func. if func passed,
     it should accepts timestamp as arg and return string namespace.
    :param ttl: expiration time for events stored, default to 1 day.
    :param socket_timeout: redis socket timeout
    :param kwargs: kwargs to be passed to redis instance init func.
    """
    def __init__(self, redis_dsn, strict=False, namespace=None, ttl=3600 * 24,
                 socket_timeout=1, **kwargs):
        super(RedisPrepareCommit, self).__init__()

        self.r = redis.StrictRedis.from_url(
            redis_dsn, socket_timeout=socket_timeout, **kwargs)
        self.strict = strict
        self.ttl = ttl
        self.logger = logging.getLogger("meepo2.prepare_commit.redis_pc")

        if namespace is None:
            self.namespace = lambda ts: "meepo2:redis_pc:%s" % d(ts, "%Y%m%d")
        elif isinstance(namespace, str):
            self.namespace = lambda ts: namespace
        elif callable(namespace):
            self.namespace = namespace

    def _keygen(self, session):
        if not hasattr(session, "meepo2_prepare_ts"):
            session.meepo2_prepare_ts = int(time.time())
        prefix = self.namespace(session.meepo2_prepare_ts)
        sp_key = "%s:session_prepare" % prefix
        sp_hkey = "%s:%s" % (sp_key, session.meepo2_unique_id)
        return sp_key, sp_hkey

    def phase(self, session):
        """Determine the session phase in prepare commit.

        :param session: sqlalchemy session
        :return: phase "prepare" or "commit"
        """
        sp_key, _ = self._keygen(session)
        if self.r.sismember(sp_key, session.meepo2_unique_id):
            return "prepare"
        else:
            return "commit"

    @_redis_strict_pc
    def prepare(self, session, event):
        """Prepare phase for session.

        :param session: sqlalchemy session
        """
        if not event:
            self.logger.warn("event empty!")
            return

        sp_key, sp_hkey = self._keygen(session)

        def _pk(obj):
            pk_values = tuple(getattr(obj, c.name)
                              for c in obj.__mapper__.primary_key)
            if len(pk_values) == 1:
                return pk_values[0]
            return pk_values

        def _get_dump_value(value):
            if hasattr(value, '__mapper__'):
                return _pk(value)
            return value
        pickled_event = {
            k: pickle.dumps({_get_dump_value(obj) for obj in objs})
            for k, objs in event.items()}
        with self.r.pipeline(transaction=False) as p:
            p.sadd(sp_key, session.meepo2_unique_id)
            p.hmset(sp_hkey, pickled_event)
            p.execute()

    @_redis_strict_pc
    def commit(self, session):
        """Commit phase for session.

        :param session: sqlalchemy session
        """
        sp_key, sp_hkey = self._keygen(session)
        with self.r.pipeline(transaction=False) as p:
            p.srem(sp_key, session.meepo2_unique_id)
            p.expire(sp_hkey, 60 * 60)
            p.execute()
    # we don't need to specially deal with rollback in this phase
    rollback = commit

    def session_info(self, session):
        """Return all session unique ids recorded in prepare phase.

        :param ts: timestamp, default to current timestamp
        :return: set of session unique ids
        """
        _, sp_hkey = self._keygen(session)
        picked_event = self.r.hgetall(sp_hkey)
        event = {s(k): pickle.loads(v) for k, v in picked_event.items()}
        return event

    def prepare_info(self, ts=None):
        """Return all session unique ids recorded in prepare phase.

        :param ts: timestamp, default to current timestamp
        :return: set of session unique ids
        """
        sp_key = "%s:session_prepare" % self.namespace(ts or int(time.time()))
        return set(s(m) for m in self.r.smembers(sp_key))

    def clear(self, ts=None):
        """Clear all session in prepare phase.

        :param ts: timestamp used locate the namespace
        """
        sp_key = "%s:session_prepare" % self.namespace(ts or int(time.time()))
        return self.r.delete(sp_key)



# -*- coding: utf-8 -*-

"""
The sqlalchemy pub will hook into SQLAlchemy's event system, shape and publish
events with ``table_action pk`` style.

The events pub flow::

    +------------------+         +-----------------------+
    |                  |         |                       |
    |    meepo2.pub     |         |      before_flush     |
    |                  |    +---->                       |
    |  sqlalchemy_pub  |    |    |  record model states  |
    |                  |    |    |                       |
    +---------+--------+    |    +-----------+-----------+
              |             |                |
        hook  |             |                |
              |             |                |
    +---------v--------+    |    +-----------v-----------+
    |                  |    |    |                       |
    |    sqlalchemy    |    |    |     after_commit      |
    |                  +----+---->                       |
    |  session events  |         |  publish model states |
    |                  |         |                       |
    +------------------+         +-----------+-----------+
                                             |
                                       +-----+
                                       |
               +------------------------------------------------+
               |                       |                        |
    +----------+--------+   +----------v---------+   +----------v---------+
    |                   |   |                    |   |                    |
    | table_write event |   | table_update event |   | table_delete event |
    |                   |   |                    |   |                    |
    +-------------------+   +--------------------+   +--------------------+

"""

from __future__ import absolute_import

import logging

import uuid

from sqlalchemy import event

from ..signals import signal


class sqlalchemy_pub(object):
    """SQLAlchemy Pub.

    The install method will add 2 hooks on sqlalchemy events system:

    * ``session_update`` -> sqlalchemy - ``before_flush``
    * ``session_commit`` -> sqlalchemy - ``after_commit``

    The ``session_update`` method need to record the model states in
    sqlalchemy "before_flush" event, when the session records the status
    with ``session.new``, ``session.dirty`` and ``session.deleted``, these
    states will be deleted in "after_commit" event.

    **General Usage**

    Install the sqlalchemy pub hook by calling it on sqlalchemy session::

        sqlalchemy_pub(session)

    Only listen some tables::

        sqlalchemy_pub(session, tables=["test"])

    Tables can be added later, the duplicated tables will be automatically
    merged::

        pub = sqlalchemy_pub(session)
        pub(["table_a", "table_b"])
        pub(["table_b", "table_c"])
        pub.tables == {"table_a", "table_b", "table_c"}

    Then use the session as usual and the events will be available.

    **Signals Illustrate**

    Sometimes you want more info than the pk value, the sqlalchemy_pub expose
    a raw signal which will send the original sqlalchemy objects.

    For example, this code::

        class Test(Base):
            __tablename__ = "test"
            id = Column(Integer, primary_key=True)
            data = Column(String)

        t_1 = Test(id=1, data='a')
        session.add(t_1)
        session.commit()

    Generates signals equal to::

        signal("test_write").send(1)
        signal("test_write_raw").send(t_1)

    :param session: sqlalchemy session to install the hook
    :param tables: tables to install the hook, leave None to pub all.

    .. warning::

        SQLAlchemy bulk operation currently **NOT** supported, so this code
        won't work::

            # bulk updates
            session.query(Test).update({"data": 'x'})

            # bulk deletes
            session.query(Test).filter(Test.data == 'x').delete()
    """

    logger = logging.getLogger("meepo2.pub.sqlalchemy_pub")

    def __init__(self, session, tables=None):
        self.session = session
        self.tables = tables or set()

        self._install()

    def __call__(self, tables):
        self.tables |= set(tables)

    def _install(self):
        # enable session_update & session_commit hook
        event.listen(self.session, "before_flush", self.session_update)
        event.listen(self.session, "after_commit", self.session_commit)

    def _pk(self, obj):
        """Get pk values from object

        :param obj: sqlalchemy object
        """
        pk_values = tuple(getattr(obj, c.name)
                          for c in obj.__mapper__.primary_key)
        if len(pk_values) == 1:
            return pk_values[0]
        return pk_values

    def _session_init(self, session):
        if hasattr(session, "meepo2_unique_id"):
            self.logger.debug("skipped - session_init")
            return

        for action in ("write", "update", "delete"):
            attr = "pending_%s" % action
            if not hasattr(session, attr):
                setattr(session, attr, set())
        session.meepo2_unique_id = uuid.uuid4().hex
        self.logger.debug("%s - session_init" % session.meepo2_unique_id)

    def _session_del(self, session):
        self.logger.debug("%s - session_del" % session.meepo2_unique_id)
        del session.meepo2_unique_id
        del session.pending_write
        del session.pending_update
        del session.pending_delete

    def _session_pub(self, session):
        def _pub(obj, action):
            """Publish object pk values with action.

            The _pub will trigger 2 signals:
            * normal ``table_action`` signal, sends primary key
            * raw ``table_action_raw`` signal, sends sqlalchemy object

            :param obj: sqlalchemy object
            :param action: action on object
            """
            if self.tables and obj.__table__.fullname not in self.tables:
                return

            sg_name = "%s_%s" % (obj.__table__, action)
            sg = signal(sg_name)
            sg_raw = signal("%s_raw" % sg_name)

            pk = self._pk(obj)
            if pk:
                sg.send(pk)
                sg_raw.send(obj)
                self.logger.debug("%s - session_pub: %s -> %s" % (
                    session.meepo2_unique_id, sg_name, pk))

        for obj in session.pending_write:
            _pub(obj, action="write")
        for obj in session.pending_update:
            _pub(obj, action="update")
        for obj in session.pending_delete:
            _pub(obj, action="delete")

        session.pending_write.clear()
        session.pending_update.clear()
        session.pending_delete.clear()

    def session_update(self, session, *_):
        """Record the sqlalchemy object states in the middle of session,
        prepare the events for the final pub in session_commit.
        """
        self._session_init(session)
        session.pending_write |= set(session.new)
        session.pending_update |= set(session.dirty)
        session.pending_delete |= set(session.deleted)
        self.logger.debug("%s - session_update" % session.meepo2_unique_id)

    def session_commit(self, session):
        """Pub the events after the session committed.

        This method should be linked to sqlalchemy "after_commit" event.
        """
        # this may happen when there's nothing to commit
        if not hasattr(session, 'meepo2_unique_id'):
            self.logger.debug("skipped - session_commit")
            return

        self._session_pub(session)
        self._session_del(session)



# -*- coding: utf-8 -*-

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import itertools
import logging

from .._compat import urlparse
from ..signals import signal


def statsd_sub(statsd_dsn, tables, prefix="meepo2.stats", rate=1):
    """statsd sub for simple stats aggregation on table events.

    This sub will send stats info to statsd daemon for counter statistics.

    To use this sub, you have to install statsd python lib with::

        $ pip install statsd

    For more info about statsd, refer to:

        https://github.com/etsy/statsd/

    :param statsd_dsn: statsd server dsn
    :param tables: table stats to be collected.
    :param prefix: statsd key prefix
    :param rate: statsd sample rate, default to 100%.
    """
    logger = logging.getLogger("meepo2.sub.statsd_sub")

    import statsd

    parsed = urlparse(statsd_dsn)
    c = statsd.StatsClient(parsed.hostname, parsed.port, prefix=prefix)

    _key = lambda tb, ac: "%s.%s" % (tb, ac)

    events = ("%s_%s" % (tb, action) for tb, action in
              itertools.product(*[tables, ["write", "update", "delete"]]))
    for event in events:
        signal(event).connect(
            lambda _, e=event: c.incr(_key(*e.rsplit('_', 1)), rate=rate),
            weak=False
        )

    logger.debug("Statsd sub hook enabled.")

    return c



# -*- coding: utf-8 -*-

from __future__ import absolute_import

import logging

from ..signals import signal


def nano_sub(bind, tables):
    """Nanomsg fanout sub. (Experimental)

    This sub will use nanomsg to fanout the events.

    :param bind: the zmq pub socket or zmq device socket.
    :param tables: the events of tables to follow.
    """
    logger = logging.getLogger("meepo2.sub.nano_sub")

    from nanomsg import Socket, PUB

    pub_socket = Socket(PUB)
    pub_socket.bind(bind)

    def _sub(table):
        for action in ("write", "update", "delete"):
            def _sub(pk, action=action):
                msg = bytes("%s_%s %s" % (table, action, pk), 'utf-8')
                logger.debug("pub msg %s" % msg)
                pub_socket.send(msg)

            signal("%s_%s" % (table, action)).connect(_sub, weak=False)

    for table in set(tables):
        _sub(table)



# -*- coding: utf-8 -*-

from __future__ import absolute_import

import collections
import zmq

from zmq.utils.strtypes import asbytes
from . import Replicator


class RqReplicator(Replicator):
    """Replicator suitable for rq task queue.

    For example:
    >>> rq_repl = RqReplicator("tcp://127.0.0.1:4000")
    >>> @rq_repl.event("a_table_update")
    >>> def job_test(pks):
    >>>     q = rq.Queue("update_cache:a_table")
    >>>     q.enqueue("module.jobs.func", pks)
    >>> rq_repl.run()

    Rq queue should be created in the external code.

    In fact this replicator can be generally used. It will pass pks as argument
    to the supplied callback func and the func can do anything you want.

    The callback func should always accept a list of primary keys.
    """
    def __init__(self, *args, **kwargs):
        super(RqReplicator, self).__init__(*args, **kwargs)

        self.topic_funcs = {}

    def event(self, *topics):
        def wrapper(func):
            for topic in topics:
                self.topic_funcs[topic] = func
                self.socket.setsockopt(zmq.SUBSCRIBE, asbytes(topic))
            return func
        return wrapper

    def run(self):
        if isinstance(self.listen, list):
            for i in self.listen:
                self.socket.connect(i)
        else:
            self.socket.connect(self.listen)

        error_pks = collections.defaultdict(set)

        def do_job(topic, pks):
            try:
                self.topic_funcs[topic](pks)
            except Exception as e:
                self.logger.exception(e)
                error_pks[topic].update(pks)
            else:
                # remove error pks
                if topic in error_pks:
                    error_pks[topic].difference_update(pks)
                    if not error_pks[topic]:
                        error_pks.pop(topic)
        try:
            while True:
                # retry error pks
                for t, p in list(error_pks.items()):
                    self.logger.warn(
                        "process error pks: {} -> {}".format(t, p))
                    do_job(t, p)

                msg = self.socket.recv_string()

                lst = msg.split()
                if len(lst) >= 2:
                    topic, pks = lst[0], lst[1:]
                else:
                    self.logger.error("msg corrupt -> %s" % msg)
                    continue

                self.logger.info("replicator: {} -> {}".format(topic, pks))
                do_job(topic, pks)
        except Exception as e:
            self.logger.exception(e)



# -*- coding: utf-8 -*-

"""
Meepo2 sub is where all the imagination comes true, all subs implemented here
are only some simple demos. Customize your own sub for the real power.

To make use of a signal, just create a function that accepts a primary key.

For example, print an event with::

    # use weak False here to force strong ref to the lambda func.
    signal("test_write").connect(
        lambda pk: logger.info("%s -> %s" % event, pk),
        weak=False
    )

For advanced use with sqlalchemy, you may also use the raw signal::

    signal("test_write_raw").connect(
        lambda obj: logger.info("%s -> %s" % event, obj.__dict__),
        weak=False)
    )
"""



# -*- coding: utf-8 -*-

"""
For basic concept about eventsourcing, refer to
http://martinfowler.com/eaaDev/EventSourcing.html

**Simple Eventsourcing**

The eventsourcing implemented in meepo2 is a simplified version of es, it only
records what has changed since a timestamp, but not the diffs.

So you only get a list of primary keys when query with a timestamp::

    order_update 102 27 59 43

Because event sourcing is hard in distributed system, you can't
give a accurate answer of events happening order. So we only keep a record
of what happened since some time, then you know the data has gone stale,
and you have to retrieve latest data from source and do the tasks upon it.

**Why Eventsourcing**

Why is eventsourcing needed? Let's check the sqlalchemy_pub events flow:

a. before flush -> record instances states
b. commit transaction in database
c. after commit -> pub signal

So it's possible that the process(or thread or greenlet) somehow being killed
right between b and c, then the signal lost.

With prepare commit in event sourcing, the session will be recorded so it's
possible to recover from this corrupt state.

But you should note this is a very rare, so in most cases, you don't need
this 100% grantee on events, then just use the simple :func:`sqlalchemy_pub`
is enough.
"""

from __future__ import absolute_import

__all__ = ["sqlalchemy_es_pub", "redis_es_sub", "statsd_es_sub"]

from .pub import sqlalchemy_es_pub
from .sub import redis_es_sub, statsd_es_sub



# -*- coding: utf-8 -*-

"""
The mysql pub will use the ``python-mysql-replication`` binlog stream as the
source of events.

The events pub flow::

                                                       +---------------------+
                                                       |                     |
                                                   +--->  table_write event  |
                                                   |   |                     |
                                                   |   +---------------------+
                                                   |
    +--------------------+      +---------------+  |
    |                    |      |               |  |   +---------------------+
    |        mysql       |      |   meepo2.pub   |  |   |                     |
    |                    +------>               +--+--->  table_update event |
    |  row-based binlog  |      |   mysql_pub   |  |   |                     |
    |                    |      |               |  |   +---------------------+
    +--------------------+      +---------------+  |
                                                   |
                                                   |   +---------------------+
                                                   |   |                     |
                                                   +--->  table_delete event |
                                                       |                     |
                                                       +---------------------+

"""

from __future__ import absolute_import

import logging
logger = logging.getLogger("meepo2.pub.mysql_pub")

import datetime
import random

import pymysqlreplication
from pymysqlreplication.row_event import (
    DeleteRowsEvent,
    UpdateRowsEvent,
    WriteRowsEvent,
)

from .._compat import urlparse, str
from ..signals import signal


def mysql_pub(mysql_dsn, tables=None, blocking=False, **kwargs):
    """MySQL row-based binlog events pub.

    **General Usage**

    Listen and pub all tables events::

        mysql_pub(mysql_dsn)

    Listen and pub only some tables events::

        mysql_pub(mysql_dsn, tables=["test"])

    By default the ``mysql_pub`` will process and pub all existing
    row-based binlog (starting from current binlog file with pos 0) and
    quit, you may set blocking to True to block and wait for new binlog,
    enable this option if you're running the script as a daemon::

        mysql_pub(mysql_dsn, blocking=True)

    The binlog stream act as a mysql slave and read binlog from master, so the
    server_id matters, if it's conflict with other slaves or scripts, strange
    bugs may happen. By default, the server_id is randomized by
    ``randint(1000000000, 4294967295)``, you may set it to a specific value
    by server_id arg::

        mysql_pub(mysql_dsn, blocking=True, server_id=1024)

    **Signals Illustrate**

    Sometimes you want more info than the pk value, the mysql_pub expose
    a raw signal which will send the original binlog stream events.

    For example, the following sql::

        INSERT INTO test (data) VALUES ('a');

    The row-based binlog generated from the sql, reads by binlog stream and
    generates signals equals to::

        signal("test_write").send(1)
        signal("test_write_raw").send({'values': {'data': 'a', 'id': 1}})

    **Binlog Pos Signal**

    The mysql_pub has a unique signal ``mysql_binlog_pos`` which contains
    the binlog file and binlog pos, you can record the signal and resume
    binlog stream from last position with it.

    :param mysql_dsn: mysql dsn with row-based binlog enabled.
    :param tables: which tables to enable mysql_pub.
    :param blocking: whether mysql_pub should wait more binlog when all
     existing binlog processed.
    :param kwargs: more kwargs to be passed to binlog stream.
    """
    # parse mysql settings
    parsed = urlparse(mysql_dsn)
    mysql_settings = {
        "host": parsed.hostname,
        "port": parsed.port or 3306,
        "user": parsed.username,
        "passwd": parsed.password
    }

    # connect to binlog stream
    stream = pymysqlreplication.BinLogStreamReader(
        mysql_settings,
        server_id=random.randint(1000000000, 4294967295),
        blocking=blocking,
        only_events=[DeleteRowsEvent, UpdateRowsEvent, WriteRowsEvent],
        **kwargs
    )

    def _pk(values):
        if isinstance(event.primary_key, str):
            return values[event.primary_key]
        return tuple(values[k] for k in event.primary_key)

    for event in stream:
        if not event.primary_key:
            continue

        if tables and event.table not in tables:
            continue

        try:
            rows = event.rows
        except (UnicodeDecodeError, ValueError) as e:
            logger.exception(e)
            continue

        timestamp = datetime.datetime.fromtimestamp(event.timestamp)

        if isinstance(event, WriteRowsEvent):
            sg_name = "%s_write" % event.table
            sg = signal(sg_name)
            sg_raw = signal("%s_raw" % sg_name)

            for row in rows:
                pk = _pk(row["values"])
                sg.send(pk)
                sg_raw.send(row)

                logger.debug("%s -> %s, %s" % (sg_name, pk, timestamp))

        elif isinstance(event, UpdateRowsEvent):
            sg_name = "%s_update" % event.table
            sg = signal(sg_name)
            sg_raw = signal("%s_raw" % sg_name)

            for row in rows:
                pk = _pk(row["after_values"])
                sg.send(pk)
                sg_raw.send(row)

                logger.debug("%s -> %s, %s" % (sg_name, pk, timestamp))

        elif isinstance(event, DeleteRowsEvent):
            sg_name = "%s_delete" % event.table
            sg = signal(sg_name)
            sg_raw = signal("%s_raw" % sg_name)

            for row in rows:
                pk = _pk(row["values"])
                sg.send(pk)
                sg_raw.send(row)

                logger.debug("%s -> %s, %s" % (sg_name, pk, timestamp))

        signal("mysql_binlog_pos").send(
            "%s:%s" % (stream.log_file, stream.log_pos))



# -*- coding: utf-8 -*-

from __future__ import absolute_import

import collections
import logging
import os
import signal
import time

from multiprocessing import Process

from ..._compat import Empty


def _deduplicate(queue, max_size):
    items = []
    for i in range(0, max_size):
        try:
            items.append(queue.get_nowait())
        except Empty:
            break
    items = set(items)
    for item in items:
        queue.put(item)


class Worker(Process):
    """Worker process

    The worker wraps the user supplied callback func. It uses multiprocessing
    queue as the task queue and will retry when a task failed.
    """
    MAX_PK_COUNT = 256

    def __init__(self, queue, name, cb, multi=False, logger_name=None,
                 retry=True, queue_limit=10000, max_retry_count=10,
                 max_retry_interval=60):
        """
        :param multi: allow multiple pks to be sent in one callback
        :param retry: retry on pk if callback failed
        :param queue_limit: queue size limit for deduplication
        :param max_retry_count: max retry count for a single pk
        :param max_retry_interval: max sleep time when callback failed
        """
        super(Worker, self).__init__()
        self.name = name
        self.queue = queue
        self.queue_limit = queue_limit
        self.cb = cb
        self.multi = multi
        self.retry = retry

        # config logger
        logger_name = logger_name or "name-%s" % id(self)
        self.logger = logging.getLogger(logger_name)
        self.logger.debug("worker %s initing..." % self.name)

        # config retry
        self._max_retry_interval = max_retry_interval
        self._max_retry_count = max_retry_count
        self._retry_stats = collections.Counter()

    def run(self):
        self.logger.debug("worker %s running..." % self.name)
        while True:
            try:
                pks = set()

                try:
                    max_size = self.queue.qsize()
                    if max_size > self.queue_limit:
                        self.logger.info("worker %s deduplicating" % self.name)
                        _deduplicate(self.queue, max_size)
                except NotImplementedError:
                    pass

                # try get all pks from queue at once
                while not self.queue.empty():
                    pks.add(self.queue.get())
                    if len(pks) > self.MAX_PK_COUNT:
                        break

                # take a nap if queue is empty
                if not pks:
                    time.sleep(1)
                    continue

                # keep order to match the results
                pks = list(pks)

                try:
                    # Mac / UNIX don't support qsize
                    self.logger.info("{0} -> {1} - qsize: {2}".format(
                        self.name, pks, self.queue.qsize()))
                except NotImplementedError:
                    self.logger.info("{0} -> {1}".format(self.name, pks))

                if self.multi:
                    results = self.cb(pks)
                else:
                    results = [self.cb(pk) for pk in pks]

                if not self.retry:
                    continue

                # check failed task and retry
                for pk, r in zip(pks, results):
                    if r:
                        self.on_success(pk)
                    else:
                        self.on_fail(pk)

                # take a nap on fail
                if not all(results):
                    time.sleep(min(3 * results.count(False),
                                   self._max_retry_interval))

            except KeyboardInterrupt:
                self.logger.debug("KeyboardInterrupt stop %s" % self.name)
                break

            except Exception as e:
                self.logger.exception(e)
                time.sleep(10)

    def on_fail(self, pk):
        self._retry_stats[pk] += 1
        if self._retry_stats[pk] > self._max_retry_count:
            del self._retry_stats[pk]
            self.logger.error("callback on pk failed -> %s" % pk)
        else:
            # put failed pk back to queue
            self.queue.put(pk)
            self.logger.warn(
                "callback on pk failed for %s times -> %s" % (
                    self._retry_stats[pk], pk))

    def on_success(self, pk):
        if pk in self._retry_stats:
            del self._retry_stats[pk]


class WorkerPool(object):
    """Manage a set of workers and recreate worker when worker dead.
    """
    def __init__(self, queues, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs
        self._queues = queues

        self._sentinel_worker = None
        self.waiting_time = kwargs.pop("waiting_time", 10)

    def _make_worker(self, queue):
        return Worker(queue, *self._args, **self._kwargs)

    def terminate(self):
        os.kill(self._sentinel_worker.pid, signal.SIGINT)
        self._sentinel_worker.join()

    def start(self):
        logger = logging.getLogger("meepo2.replicator.sentinel")

        def _f():
            worker_map = {
                q: self._make_worker(q) for q in self._queues
            }
            for _, worker in worker_map.items():
                worker.start()

            logger.info("starting sentinel...")
            try:
                while True:
                    logger.debug("ping {} worker".format(self._args[0]))
                    dead = qsize = 0
                    for queue, worker in worker_map.items():
                        try:
                            qsize += queue.qsize()
                        except NotImplementedError:
                            qsize = None

                        if not worker.is_alive():
                            dead += 1
                            logger.warn(
                                "{} worker {} dead, recreating...".format(
                                    self._args[0], worker.pid))

                            worker_map[queue] = self._make_worker(queue)
                            worker_map[queue].start()

                    msg = ["{} total qsize {}".format(self._args[0], qsize),
                           "{} worker alive, {} worker dead".format(
                               len(worker_map) - dead, dead)]

                    logger.info("; ".join(msg))

                    time.sleep(self.waiting_time)
            except KeyboardInterrupt:
                pass
            finally:
                for worker in worker_map.values():
                    worker.terminate()

        self._sentinel_worker = Process(target=_f)
        self._sentinel_worker.start()



# -*- coding: utf-8 -*-

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import logging

import zmq


def zmq_proxy(frontend, backend):
    logger = logging.getLogger("meepo2.replicator.zmq_proxy")

    ctx, frontend_socket, backend_socket = None, None, None

    try:
        ctx = zmq.Context()
        frontend_socket = ctx.socket(zmq.XSUB)
        frontend_socket.bind(frontend)

        backend_socket = ctx.socket(zmq.XPUB)
        backend_socket.bind(backend)

        zmq.proxy(frontend_socket, backend_socket)

    except Exception as e:
        logger.exception(e)

    finally:
        if frontend_socket:
            frontend_socket.close()
        if backend_socket:
            backend_socket.close()
        if ctx:
            ctx.term()



# -*- coding: utf-8 -*-

from __future__ import absolute_import

import logging
import time

import redis

from ...utils import s, d


class EventStore(object):
    def __init__(self):
        pass

    def add(self, event, pk, ts=None):
        raise NotImplementedError

    def replay(self, event, ts=0, end_ts=None, with_ts=False):
        raise NotImplementedError

    def query(self, event, pk, ts=None):
        raise NotImplementedError

    def clear(self, event, ts=None):
        raise NotImplementedError


class RedisEventStore(EventStore):
    """EventStore based on redis.

    The event store use namespace and event name as key and store primary
    keys using redis sorted set, with event timestamp as score.

    **General Usage**

    Init event store with redis_dsn::

        event_store = RedisEventStore("redis://localhost/", "store")

    You can also pass a function to namespace, it'll accept timestamp as
    arg, this can be used to separate events store based on hour, day or
    week etc.::

        event_store = RedisEventStore(
            "redis://localhost/", lambda ts: "store:%s" % d(ts, "%Y%m%d"))

    Add a event with::

        event_store.add("test_write", 1)

    Or add a event with timestamp passed in::

        event_store.add("test_write", 2, ts=1024)

    Clear all records of an event within a namespace::

        event_store.clear("test_write")

    **Events Replay**

    One important feature for eventsourcing is replay, it can replay what has
    changed and the latest update timestamp of events.

    Replay all records of an event within a namespace::

        event_store.replay("test_write")

    Or replay all records since timestamp::

        # all events since timestamp 1024
        event_store.replay("test_write", ts=1024)

        # all events between timestamp 1024 and now
        event_store.replay("test_write", ts=1024, end_ts=time.time())

    You can also replay all events with it's latest updating time::

        event_store.replay("test_write", with_ts=True)

    **Events Query**

    You can query the last change timestamp of an event with query api.

    Query records within current namespace::

        event_store.query("test_write", 1)

    The return value will either be int timestamp or None if record not
    exists.

    Add a timestamp to query events within other namespace (assume you
    separate the event store namespace by day, you may want to query event
    happened yesterday.)::

        event_store.query("test_write", 1, ts=some_value)

    .. note::

        The redis event store class is compat with twemproxy.

    :param redis_dsn: the redis instance uri
    :param namespace: namespace func for event key, the func should accept
     event timestamp and return namespace of the func. namespace also
     accepts str type arg, which will always return the same namespace
     for all timestamps.
    :param ttl: expiration time for events stored, default to 3 days.
    :param socket_timeout: redis socket timeout
    :param kwargs: kwargs to be passed to redis instance init func.
    """

    LUA_TIME = "return tonumber(redis.call('TIME')[1])"
    LUA_ZADD = ' '.join("""
    local score = redis.call('ZSCORE', KEYS[1], ARGV[2])
    if score and tonumber(ARGV[1]) <= tonumber(score) then
        return 0
    else
        redis.call('ZADD', KEYS[1], ARGV[1], ARGV[2])
        return 1
    end
    """.split())

    def __init__(self, redis_dsn, namespace=None, ttl=3600 * 24 * 3,
                 socket_timeout=1, **kwargs):
        super(RedisEventStore, self).__init__()

        self.r = redis.StrictRedis.from_url(
            redis_dsn, socket_timeout=socket_timeout, **kwargs)
        self.ttl = ttl
        self.logger = logging.getLogger("meepo2.redis_es")

        if namespace is None:
            self.namespace = lambda ts: "meepo2:redis_es:%s" % d(ts, "%Y%m%d")
        elif isinstance(namespace, str):
            self.namespace = lambda ts: namespace
        elif callable(namespace):
            self.namespace = namespace

    def _keygen(self, event, ts=None):
        """Generate redis key for event at timestamp.

        :param event: event name
        :param ts: timestamp, default to current timestamp if left as None
        """
        return "%s:%s" % (self.namespace(ts or time.time()), event)

    def _time(self):
        """Redis lua func to get timestamp from redis server, use this func to
        prevent time inconsistent across servers.
        """
        return self.r.eval(self.LUA_TIME, 1, 1)

    def _zadd(self, key, pk, ts=None, ttl=None):
        """Redis lua func to add an event to the corresponding sorted set.

        :param key: the key to be stored in redis server
        :param pk: the primary key of event
        :param ts: timestamp of the event, default to redis_server's
         current timestamp
        :param ttl: the expiration time of event since the last update
        """
        return self.r.eval(self.LUA_ZADD, 1, key, ts or self._time(), pk)

    def add(self, event, pk, ts=None, ttl=None):
        """Add an event to event store.

        All events were stored in a sorted set in redis with timestamp as
        rank  score.

        :param event: the event to be added, format should be ``table_action``
        :param pk: the primary key of event
        :param ts: timestamp of the event, default to redis_server's
         current timestamp
        :param ttl: the expiration time of event since the last update
        :return: bool
        """
        key = self._keygen(event, ts)
        try:
            self._zadd(key, pk, ts, ttl)
            return True
        except redis.ConnectionError as e:
            # connection error typically happens when redis server can't be
            # reached or timed out, the error will be silent with an error
            # log and return None.
            self.logger.error(
                "redis event store failed with connection error %r" % e)
            return False

    def replay(self, event, ts=0, end_ts=None, with_ts=False):
        """Replay events based on timestamp.

        If you split namespace with ts, the replay will only return events
        within the same namespace.

        :param event: event name
        :param ts: replay events after ts, default from 0.
        :param end_ts: replay events to ts, default to "+inf".
        :param with_ts: return timestamp with events, default to False.
        :return: list of pks when with_ts set to False, list of (pk, ts) tuples
         when with_ts is True.
        """
        key = self._keygen(event, ts)
        end_ts = end_ts if end_ts else "+inf"
        elements = self.r.zrangebyscore(key, ts, end_ts, withscores=with_ts)

        if not with_ts:
            return [s(e) for e in elements]
        else:
            return [(s(e[0]), int(e[1])) for e in elements]

    def query(self, event, pk, ts=None):
        """Query the last update timestamp of an event pk.

        You can pass a timestamp to only look for events later than that
        within the same namespace.

        :param event: the event name.
        :param pk: the pk value for query.
        :param ts: query event pk after ts, default to None which will query
         all span of current namespace.
        """
        key = self._keygen(event, ts)
        pk_ts = self.r.zscore(key, pk)
        return int(pk_ts) if pk_ts else None

    def clear(self, event, ts=None):
        """Clear all stored record of event.

        :param event: event name to be cleared.
        :param ts: timestamp used locate the namespace
        """
        return self.r.delete(self._keygen(event, ts))



# -*- coding: utf-8 -*-

from __future__ import absolute_import

import datetime
import functools
import itertools
import logging

from ..._compat import urlparse
from ...signals import signal

from .event_store import RedisEventStore
from .prepare_commit import RedisPrepareCommit


def redis_es_sub(session, tables, redis_dsn, strict=False,
                 namespace=None, ttl=3600 * 24 * 3, socket_timeout=1):
    """Redis EventSourcing sub.

    This sub should be used together with sqlalchemy_es_pub, it will
    use RedisEventStore as events storage layer and use the prepare-commit
    pattern in :func:`sqlalchemy_es_pub` to ensure 100% security on
    events recording.

    :param session: the sqlalchemy to bind the signal
    :param tables: tables to be event sourced.
    :param redis_dsn: the redis server to store event sourcing events.
    :param strict: arg to be passed to RedisPrepareCommit. If set to True,
     the exception will not be silent and may cause the failure of sqlalchemy
     transaction, user should handle the exception in the app side in this
     case.
    :param namespace: namespace string or func. If func passed, it should
     accept timestamp as arg and return a string namespace.
    :param ttl: expiration time for events stored, default to 3 days.
    :param socket_timeout: redis socket timeout.
    """
    logger = logging.getLogger("meepo2.sub.redis_es_sub")

    if not isinstance(tables, (list, set)):
        raise ValueError("tables should be list or set")

    # install event store hook for tables
    event_store = RedisEventStore(
        redis_dsn, namespace=namespace, ttl=ttl, socket_timeout=socket_timeout)

    def _es_event_sub(pk, event):
        if event_store.add(event, str(pk)):
            logger.info("%s: %s -> %s" % (
                event, pk, datetime.datetime.now()))
        else:
            logger.error("event sourcing failed: %s" % pk)

    events = ("%s_%s" % (tb, action) for tb, action in
              itertools.product(*[tables, ["write", "update", "delete"]]))
    for event in events:
        sub_func = functools.partial(_es_event_sub, event=event)
        signal(event).connect(sub_func, weak=False)

    # install prepare-commit hook
    prepare_commit = RedisPrepareCommit(
        redis_dsn, strict=strict, namespace=namespace,
        socket_timeout=socket_timeout)

    signal("session_prepare").connect(
        prepare_commit.prepare, sender=session, weak=False)
    signal("session_commit").connect(
        prepare_commit.commit, sender=session, weak=False)
    signal("session_rollback").connect(
        prepare_commit.rollback, sender=session, weak=False)

    logger.debug("Redis event sourcing hook enabled.")

    return event_store, prepare_commit


def statsd_es_sub(statsd_dsn, session, prefix="meepo2.stats", rate=1):
    """statsd sub for simple stats aggregation on event sourcing.

    This sub will send stats info to statsd daemon for counter statistics,
    this one is different from `meepo2.sub.statsd_sub` which collect stats
    for table events, the `statsd_es_sub` only collects statsd for session
    events (prepare, commit, rollback).

    You may wish to use this sub in combination of `meepo2.sub.statsd_sub`
    when used in event sourcing.

    :param statsd_dsn: statsd server dsn
    :param session: session to collect the stats.
    :param prefix: statsd key prefix
    :param rate: statsd sample rate, default to 100%.
    """
    logger = logging.getLogger("meepo2.sub.statsd_es_sub")

    import statsd

    parsed = urlparse(statsd_dsn)
    c = statsd.StatsClient(parsed.hostname, parsed.port, prefix=prefix)

    _incr = lambda p: c.incr("session_%s" % p, rate=rate)

    signal("session_prepare").connect(
        lambda _, event: _incr("prepare"), sender=session, weak=False
    )
    signal("session_commit").connect(
        lambda _: _incr("commit"), sender=session, weak=False
    )
    signal("session_rollback").connect(
        lambda _: _incr("rollback"), sender=session, weak=False
    )

    logger.debug("Statsd event sourcing sub hook enabled.")

    return c



# -*- coding: utf-8 -*-

from __future__ import absolute_import


__all__ = ["pickle", "urlparse", "Empty"]

import sys
PY3 = sys.version_info[0] >= 3

if PY3:
    from urllib.parse import urlparse
    from queue import Empty
    import pickle

    bytes = bytes
    str = str

else:
    from urlparse import urlparse
    from Queue import Empty
    import cPickle as pickle

    bytes = str
    str = unicode  # noqa



# -*- coding: utf-8 -*-

from __future__ import absolute_import

import logging

import collections

from sqlalchemy import event

from ...pub.sqlalchemy import sqlalchemy_pub
from ...signals import signal


class sqlalchemy_es_pub(sqlalchemy_pub):
    """SQLAlchemy EventSourcing Pub.

    Add eventsourcing to sqlalchemy_pub, three more signals added for tables:

    * ``session_prepare``
    * ``session_commit`` / ``session_rollback``

    The hook will use prepare-commit pattern to ensure 100% reliability on
    event sourcing.

    **Multi-Sessions Prepare Commit**

    The 3 additional signals were attached to sqlalchemy session factory in
    case of being used in multi-sessions environments.

    If you only use one sqlalchemy session in your program, it's fine to use
    ``session_prepare`` / ``session_commit`` as other signals.

    But if you use multiple sessions, you can separate the prepare-commit
    signals by:

    * Separate sessions by settings ``info`` arg in session factory.

      Because the info is the only attributes copied from session factory
      to session instance.

      ``meepo2.signals`` monkey patched the blinker ``hashable_identity``
      func to use the ``session.info`` for session hash.

    * Provide session as sender when signal receivers connects.

    For example::

        # setting `info` in sqlalchemy session_factory
        SessionA = sessionmaker(bind=engine_a, info={"name": "session_a"})
        SessionB = sessionmaker(bind=engine_b, info={"name": "session_b"})

        sqlalchemy_es_pub(SessionA)
        sqlalchemy_es_pub(SessionB)

        sg = signal("session_prepare")

        def _sp_for_a(session, event):
            print(session.info)
        sg.connect(_sp_for_a, sender=SessionA)

    Then the ``_sp_for_a`` will only receive prepare-commit related events
    triggered by ``SessionA``.
    """
    logger = logging.getLogger("meepo2.pub.sqlalchemy_es_pub")

    def _install(self):
        event.listen(self.session, "before_flush", self.session_update)

        # enable session prepare-commit hook
        event.listen(self.session, "after_flush", self.session_prepare)
        event.listen(self.session, "after_commit", self.session_commit)
        event.listen(self.session, "after_rollback", self.session_rollback)

    def session_prepare(self, session, _):
        """Send session_prepare signal in session "before_commit".

        The signal contains another event argument, which records whole info
        of what's changed in this session, so the signal receiver can receive
        and record the event.
        """
        if not hasattr(session, 'meepo2_unique_id'):
            self._session_init(session)

        evt = collections.defaultdict(set)
        for action in ("write", "update", "delete"):
            objs = getattr(session, "pending_%s" % action)
            # filter tables if possible
            if self.tables:
                objs = [o for o in objs
                        if o.__table__.fullname in self.tables]
            for obj in objs:
                evt_name = "%s_%s" % (obj.__table__.fullname, action)
                evt[evt_name].add(obj)
                self.logger.debug("%s - session_prepare: %s -> %s" % (
                    session.meepo2_unique_id, evt_name, self._pk(obj)))

        # only trigger signal when event exists
        if evt:
            signal("session_prepare").send(session, event=evt)

    def session_commit(self, session):
        """Send session_commit signal in sqlalchemy ``before_commit``.

        This marks the success of session so the session may enter commit
        state.
        """
        # this may happen when there's nothing to commit
        if not hasattr(session, 'meepo2_unique_id'):
            self.logger.debug("skipped - session_commit")
            return

        # normal session pub
        self.logger.debug("%s - session_commit" % session.meepo2_unique_id)
        self._session_pub(session)
        signal("session_commit").send(session)
        self._session_del(session)

    def session_rollback(self, session):
        """Send session_rollback signal in sqlalchemy ``after_rollback``.

        This marks the failure of session so the session may enter commit
        phase.
        """
        # this may happen when there's nothing to rollback
        if not hasattr(session, 'meepo2_unique_id'):
            self.logger.debug("skipped - session_rollback")
            return

        # del session meepo2 id after rollback
        self.logger.debug("%s - after_rollback" % session.meepo2_unique_id)
        signal("session_rollback").send(session)
        self._session_del(session)



# -*- coding: utf-8 -*-

from __future__ import absolute_import

import ketama
import zmq

from multiprocessing import Queue
from zmq.utils.strtypes import asbytes

from . import Replicator
from .worker import WorkerPool


class QueueReplicator(Replicator):
    """Replicator using Queue as worker task queue

    This Replicator receives events from upstream zmq devices and put them into
    a set of python multiprocessing queues using ketama consistent hashing.
    Each queue has a worker. We use :class:`WorkerPool` to manage a set of
    queues.
    """

    def __init__(self, *args, **kwargs):
        super(QueueReplicator, self).__init__(*args, **kwargs)

        # init workers
        self.workers = {}
        self.worker_queues = {}

    def event(self, *topics, **kwargs):
        """Topic callback registry.

        callback func should receive two args: topic and pk, and then process
        the replication job.

        Note: The callback func must return True/False. When passed a list of
        pks, the func should return a list of True/False with the same length
        of pks.

        :param topics: a list of topics
        :param workers: how many workers to process this topic
        :param multi: whether pass multiple pks
        :param queue_limit: when queue size is larger than the limit,
         the worker should run deduplicate procedure
        """
        workers = kwargs.pop("workers", 1)
        multi = kwargs.pop("multi", False)
        queue_limit = kwargs.pop("queue_limit", 10000)

        def wrapper(func):
            for topic in topics:
                queues = [Queue() for _ in range(workers)]
                hash_ring = ketama.Continuum()
                for q in queues:
                    hash_ring[str(hash(q))] = q
                self.worker_queues[topic] = hash_ring
                self.workers[topic] = WorkerPool(
                    queues, topic, func, multi=multi, queue_limit=queue_limit,
                    logger_name="%s.%s" % (self.name, topic))
                self.socket.setsockopt(zmq.SUBSCRIBE, asbytes(topic))
            return func
        return wrapper

    def run(self):
        """Run the replicator.

        Main process receive messages and distribute them to worker queues.
        """
        for worker_pool in self.workers.values():
            worker_pool.start()

        if isinstance(self.listen, list):
            for i in self.listen:
                self.socket.connect(i)
        else:
            self.socket.connect(self.listen)

        try:
            while True:
                msg = self.socket.recv_string()
                lst = msg.split()
                if len(lst) == 2:
                    topic, pks = lst[0], [lst[1], ]
                elif len(lst) > 2:
                    topic, pks = lst[0], lst[1:]
                else:
                    self.logger.error("msg corrupt -> %s" % msg)
                    continue

                self.logger.debug("replicator: {0} -> {1}".format(topic, pks))
                for pk in pks:
                    self.worker_queues[topic][str(hash(pk))].put(pk)
        except Exception as e:
            self.logger.exception(e)
        finally:
            for worker_pool in self.workers.values():
                worker_pool.terminate()



# -*- coding: utf-8 -*-

from __future__ import absolute_import

from logging.config import dictConfig
import datetime

from ._compat import bytes, str


def setup_logger(level=None):
    dictConfig({
        'version': 1,
        'disable_existing_loggers': False,

        'root': {
            'handlers': ['console'],
            'level': 'INFO',
        },

        'loggers': {
            'meepo2': {
                'handlers': ['console'],
                'propagate': False,
                'level': level or 'INFO',
            },
        },

        'handlers': {
            'console': {
                'level': 'DEBUG',
                'class': 'logging.StreamHandler',
                'formatter': 'console'
            },
        },

        'formatters': {
            'console': {
                'format': '%(asctime)s [%(levelname)s] [%(name)s][%(process)d]'
                          ': %(message)s',
            },
        }
    })


def cast_bytes(s, encoding='utf8', errors='strict'):
    """cast str or bytes to bytes"""
    if isinstance(s, bytes):
        return s
    elif isinstance(s, str):
        return s.encode(encoding, errors)
    else:
        raise TypeError("Expected unicode or bytes, got %r" % s)


b = cast_bytes


def cast_str(s, encoding='utf8', errors='strict'):
    """cast bytes or str to str"""
    if isinstance(s, bytes):
        return s.decode(encoding, errors)
    elif isinstance(s, str):
        return s
    else:
        raise TypeError("Expected unicode or bytes, got %r" % s)


s = cast_str


def cast_datetime(ts, fmt=None):
    """cast timestamp to datetime or date str"""
    dt = datetime.datetime.fromtimestamp(ts)
    if fmt:
        return dt.strftime(fmt)
    return dt


d = cast_datetime



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

# requirements
install_requires = [
    "SQLAlchemy>=0.9.0,<1.1.0",
    "blinker>=1.3,<2.0",
    "mysql-replication>=0.5,<0.6.0",
    "pyketama>=0.2.0",
    "pyzmq>=14.3.1,<15.0.0",
    "redis>=2.10.3,<2.11.0",
]

dev_requires = [
    "flake8>=2.2",
    "pytest>=2.6",
    "sphinx-rtd-theme>=0.1.6",
    "sphinx>=1.2",
    "tox>=1.8",
] + install_requires


setup(name="meepo2",
      version=__import__("meepo2").__version__,
      description="event sourcing for databases.",
      keywords="eventsourcing event sourcing replication cache elasticsearch",
      author="Lx Yu",
      author_email="i@lxyu.net",
      packages=find_packages(exclude=['docs', 'tests']),
      url="https://github.com/huskar-org/meepo2",
      license="MIT",
      zip_safe=False,
      long_description=open("README.rst").read(),
      install_requires=install_requires,
      extras_require={
          "dev": dev_requires,
      },
      classifiers=[
          "Topic :: Software Development",
          "Development Status :: 3 - Alpha",
          "Intended Audience :: Developers",
          "License :: OSI Approved :: MIT License",
          "Programming Language :: Python :: 2.7",
          "Programming Language :: Python :: 3.3",
          "Programming Language :: Python :: 3.4",
          "Programming Language :: Python :: Implementation :: CPython",
          "Programming Language :: Python :: Implementation :: PyPy",
      ])



# -*- coding: utf-8 -*-

from __future__ import absolute_import

import itertools
import logging

from ..signals import signal


def print_sub(tables):
    """Dummy print sub.

    :param tables: print events of tables.
    """
    logger = logging.getLogger("meepo2.sub.print_sub")
    logger.info("print_sub tables: %s" % ", ".join(tables))

    if not isinstance(tables, (list, set)):
        raise ValueError("tables should be list or set")

    events = ("%s_%s" % (tb, action) for tb, action in
              itertools.product(*[tables, ["write", "update", "delete"]]))
    for event in events:
        signal(event).connect(
            lambda pk: logger.info("%s -> %s", event, pk), weak=False)



# -*- coding: utf-8 -*-

"""
Meepo2's core concept is based event pubs, it follows mysql row-based binlog
and sqlalchemy events system and shape them into ``table_action pk``
format signals.

Currently there are 2 pubs implemented: ``mysql_pub`` and ``sqlalchemy_pub``.

The publishers and subscribers are connected with ``blinker.signal``.

Publisher sends pk by::

    signal("table_action").send(pk)

And subscriber can accept the pk by::

    sg = signal("table_action")

    @sg.connect
    def dummy_print(pk):
        print(pk)
"""

from __future__ import absolute_import

__all__ = ["mysql_pub", "sqlalchemy_pub"]

from .mysql import mysql_pub
from .sqlalchemy import sqlalchemy_pub
